# -*- coding: utf-8 -*-
"""
These transformers can be used if filenames are used as documents.
"""

import os
import pickle
import json

from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin


class FileReader(BaseEstimator, TransformerMixin):
    """
    Transforms filenames into their content.
    By default, this transformer will open files in text mode.
    This is not a generator, so be careful with your memory.

    input: list of paths
    output: list of contents of the files specified by the paths.
    """

    def __init__(self, mode='text'):
        """
        Args:
            mode: The mode in which the files are to be opened. One of 'text',
                'pickle' or 'json'.
        """
        valid_modes = ['text', 'pickle', 'json']
        if mode not in valid_modes:
            raise ValueError(f'unknown mode {mode}, use one of: {valid_modes}')
        self.mode = mode

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        ret = []
        for document in X:
            if not os.path.isfile(document):
                raise ValueError(f'No such file: {document}')
            if self.mode == 'text':
                with open(document, 'r') as i_f:
                    ret.append(i_f.read())
            elif self.mode == 'json':
                with open(document, 'r') as i_f:
                    ret.append(json.load(i_f))
            if self.mode == 'pickle':
                with open(document, 'rb') as i_f:
                    ret.append(pickle.load(i_f))
        return ret


class FileWriter(BaseEstimator, TransformerMixin):
    """
    Stores elements to filenames, which must be provided at object creation
    time.

    input: list of serializable content
    output: the same content. Storing the files is a side-effect.
    """

    def __init__(self, filenames, writemode='text'):
        """
        Args:
            filenames: one valid path for each entry in X passed to fit()
            writemode: mode in which to write the files. Can be 'text',
                'pickle' or 'pickle'.
        """

        self.filenames = filenames
        valid_writemodes = ['text', 'json', 'pickle']
        if writemode not in valid_writemodes:
            raise ValueError(f'writemode must be one of {valid_writemodes}')
        self.writemode = writemode

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        X = list(X)
        assert len(X) == len(self.filenames)
        for i, (filename, document) in enumerate(zip(self.filenames, X)):
            if not os.path.isdir(os.path.dirname(filename)):
                os.makedirs(os.path.dirname(filename), exist_ok=True)
            if self.writemode == 'text':
                with open(filename, 'w') as o_f:
                    o_f.write(document)
            elif self.writemode == 'json':
                with open(filename, 'w') as o_f:
                    json.dump(document, o_f)
            elif self.writemode == 'pickle':
                with open(filename, 'wb') as o_f:
                    pickle.dump(document, o_f)
        return X


class PathTransformer(TransformerMixin, BaseEstimator):
    """
    Transforms the path of files and allows to replace the extension.

    input: list of paths
    output: list of the same paths that have been modified according to
    the initialization parameters
    """

    def __init__(self, path, extension=None):
        """
        Args:
            path: The path which is appended to each filename
            extension: The new extensions that is used, including the dot.
                This replaces all old extensions: given a file foobar.txt.gz
                and an extension .grammar will result in foobar.grammar.
        """

        self.path = path
        self.extension = extension

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        ret = []
        for x in X:
            base = os.path.dirname(x)
            new_base = os.path.join(base, self.path)
            name = os.path.splitext(os.path.basename(x))[0]
            if self.extension:
                ret.append(os.path.join(new_base, name + self.extension))
            else:
                ret.append(os.path.join(new_base, os.path.basename(x)))
        return ret



"""Base classes for the dbispipeline."""
from abc import ABC
from abc import abstractmethod

from sklearn.model_selection import ParameterGrid


class Loader(ABC):
    """Abstract base class of a dataloader."""

    @abstractmethod
    def load(self):
        """Returns the data loaded by the dataloader."""
        pass

    @property
    @abstractmethod
    def configuration(self):
        """Returns a dict-like representation of the configuration."""
        pass

    @property
    def run_count(self):
        """Returns how many configurations this dataloader will produce."""
        return 0

    @property
    def is_multiloader(self):
        """Returns if this dataloader will produce multiple configurations."""
        return self.run_count > 0


class MultiLoaderGenerator(Loader):
    """Wrapper for loaders with parameters."""

    def __init__(self, loader_class, parameters):
        """
        Dynamically constructs a multiloader from a loader and parameters.

        Args:
            loader_class: the class of the dataloader to be instantiated.
                Do not pass an instance.
            parameters: if passed a list, this generator will return one
                dataloader instance for each entry in the list, and each entry
                in the list is passed to the constructor of the loader. If
                passed a dict, a grid of all combinations is generated and
                passed to the loader.
        """
        self.loaders = []
        if isinstance(parameters, dict):
            for sample in ParameterGrid(parameters):
                # this produces only dicts
                self.loaders.append(loader_class(**sample))
        else:
            for sample in parameters:
                if isinstance(sample, dict):
                    self.loaders.append(loader_class(**sample))
                else:
                    self.loaders.append(loader_class(*sample))

    def load(self):
        """Loads the next data configuration as a generator."""
        for loader in self.loaders:
            yield loader.load()

    @property
    def configuration(self):
        """Returns the configuration of the next data configuration."""
        for i, loader in enumerate(self.loaders):
            config = loader.configuration
            config['run_number'] = i
            config['class'] = loader.__class__.__name__
            yield config

    @property
    def run_count(self):
        """Returns how many configurations this loader will generate."""
        return len(self.loaders)


class TrainTestLoader(Loader):
    """Abstract dataloader destinguishing trian and test data."""

    @abstractmethod
    def load_train(self):
        """Returns the train data."""
        pass

    @abstractmethod
    def load_test(self):
        """Returns the test data."""
        pass

    def load(self):
        """Returns the data as a train, test tuple."""
        return self.load_train(), self.load_test()


class TrainValidateTestLoader(TrainTestLoader):
    """Abstract dataloader destinguishing trian and test data."""

    @abstractmethod
    def load_validate(self):
        """Returns the validation data."""
        pass

    def load(self):
        """Returns the data as a train, validation, test tuple."""
        return self.load_train(), self.load_validate(), self.load_test()


class Evaluator(ABC):
    """Abstract base class of an evaluator."""

    @abstractmethod
    def evaluate(self, model, data):
        """
        Evaluates the pipline based on the given dataset.

        Args:
            model: the model given in the pipeline.
            dataloader: the data used to load the dataset.

        Returns: An dict result.
        """
        pass

    @property
    @abstractmethod
    def configuration(self):
        """Returns a dict-like representation of this loader.

        This is for storing its state in the database.
        """
        pass

    def _check_loader_methods(self, loader, methods=None):

        if type(methods) == str:
            methods = [methods]

        for method in methods:
            m = getattr(loader, method, None)
            if not callable(m):
                raise ValueError(
                    f'The dataloader {loader} does not implement'
                    'the required method {m}. Required methods are: '
                    '{methods}')


class StorageHandler(ABC):
    """Abstract base class of a storage handler."""

    @abstractmethod
    def handle_result(self, result):
        """
        Handles the result of an evaluator.

        Args:
            result: passed from an evaluator.
        """
        pass



"""Core module of the dbis pipline."""

from collections.abc import Iterable
import importlib.util
import platform
import socket
import time

import git

from . import store
from .base import Loader
from .evaluators import Evaluator
from .storage_handlers import PickleBackupHandler
from .storage_handlers import PostGresOutputHandler
from .storage_handlers import StorageHandler
from .utils import LOGGER
from .utils import check_serializability
from .utils import get_backup_dir
from .utils import get_job_id
from .utils import notify_error


class Core:
    """Represents the core functionality of the DBIS pipeline."""

    def __init__(self, plan, dryrun=False, mail=None):
        """
        Intitializes the pipeline.

        Args:
            plan: either the path to or the contents of the file containing
                information about the steps that should be performed. This must
                include a valid dataloader, pipeline and evaluator object.

            dryrun: if true, this run won't be stored to the database

            mail: one of [None, 'total', 'runs'] - defines how verbose the mail
                notification system should be. None: no mails are sent. total:
                one mail is sent at the end of the entire run. runs: one mail
                is sent for every configuration generated by the dataloader.
        """
        self.dryrun = dryrun
        self.mail = mail
        self.plan_path = None
        self.plan_content = None
        self.git_info = self._get_git_info()

        if type(plan) == str:
            with open(plan, 'r') as plan_file:
                self.plan_content = plan_file.read()
            self.plan_path = plan
            plan = load_plan(plan)

        store['plan_path'] = self.plan_path
        self.dataloader = plan.dataloader
        self.pipeline = plan.pipeline
        self.evaluator = plan.evaluator
        self.storage_handlers = plan.storage_handlers
        if not dryrun:
            self.storage_handlers.insert(
                0,
                PickleBackupHandler(directory=get_backup_dir()),
            )
            self.storage_handlers.append(PostGresOutputHandler())
        if 'result_handlers' in dir(plan):
            self.result_handlers = plan.result_handlers
        else:
            self.result_handlers = []
        self.times = {}

    def run(self):
        """Runs the evaluation specified by the pipeline."""
        self.times['global_start'] = time.perf_counter()
        try:
            data = self.dataloader.load()
            loader_configs = self.dataloader.configuration
        except Exception as e:
            LOGGER.error('Error during data loading:')
            LOGGER.exception(e)
            # this is not recoverable.
            return

        if not self.dataloader.is_multiloader:
            data = [data]
            loader_configs = [loader_configs]
        for i, (run_data,
                loader_config) in enumerate(zip(data, loader_configs)):
            try:
                result = self.evaluator.evaluate(self.pipeline, run_data)
            except Exception as e:
                LOGGER.exception('Error during evaluation: %s', e)
                # this is not recoverable.
                return
            for handler in self.result_handlers:
                self._run_result_handler(result, handler, i, loader_config)
            self._store(i, result, loader_config)

    def _run_result_handler(self, result, handler, run_number, loader_config):
        try:
            if callable(handler):
                handler(result)
            else:
                function, kwargs = handler
                function(result, **kwargs)
        except Exception as e:
            if hasattr(handler, '__name__'):
                LOGGER.error('Result handler %s failed:', handler.__name__)
                LOGGER.exception(e)
            else:
                LOGGER.error('Result handler %s failed:', handler)
                LOGGER.exception(e)
            if self.mail:
                notify_error(self.plan_path,
                             'result handling',
                             e,
                             run=run_number,
                             loader_config=loader_config)

    def _get_pipeline_configuration(self, pipeline):
        clazz = pipeline.__class__.__name__

        # for custom models with detailed information
        if hasattr(pipeline, 'configuration'):
            return pipeline.configuration

        # for pipeline objects
        if hasattr(pipeline, 'named_steps'):
            steps = pipeline.named_steps.items()
            ret = {
                name: self._get_pipeline_configuration(transformer)
                for name, transformer in steps
            }
            ret['TYPE'] = clazz
            return ret

        # for featureunion, transformerpicker objects
        if hasattr(pipeline, 'transformer_list'):
            transformers = pipeline.transformer_list
            ret = {
                transformer[0]:
                    self._get_pipeline_configuration(transformer[1])
                for transformer in list(transformers)
            }
            ret['TYPE'] = clazz
            return ret

        return clazz

    def _get_git_info(self):
        sha_commit_id = None
        remote_url = None
        is_dirty = None
        try:
            repo = git.Repo(search_parent_directories=True)
            is_dirty = repo.is_dirty()
            sha_commit_id = repo.head.object.hexsha
            if repo.remotes and hasattr(repo.remotes, 'origin'):
                remote_url = repo.remotes.origin.url
            return sha_commit_id, remote_url, is_dirty
        except git.GitError:
            return None, None, None

    def _get_platform_info(self):
        info = {}
        info['arch'] = platform.architecture()
        info['machine'] = platform.machine()
        info['node'] = socket.getfqdn()
        info['platform'] = platform.platform()
        info['processor'] = platform.processor()
        info['system'] = platform.system()
        info['job_id'] = get_job_id()

        return info

    def _store(self, run_number, result, dl_conf):
        """Stores the results of the pipeline."""
        try:
            if self.dataloader.is_multiloader and 'run_number' not in dl_conf:
                dl_conf['run_number'] = run_number
            if 'class' not in dl_conf:
                dl_conf['class'] = self.dataloader.__class__.__name__

            ev_conf = self.evaluator.configuration
            if type(ev_conf) == dict:
                ev_conf['class'] = self.evaluator.__class__.__name__

            pipeline_conf = self._get_pipeline_configuration(self.pipeline)

            config = {
                'git_info': self.git_info,
                'sourcefile': self.plan_path,
                'plan_content': self.plan_content,
                'dataloader': dl_conf,
                'pipeline': pipeline_conf,
                'evaluator': ev_conf,
                'outcome': result,
                'durations': self.times,
                'platform': self._get_platform_info(),
            }

            config = check_serializability(config)

            for storage_handler in self.storage_handlers:
                try:
                    storage_handler.handle_result(config)
                except Exception as e:
                    LOGGER.error('storage handler %s failed:', storage_handler)
                    LOGGER.exception(e)
        except Exception as e:
            LOGGER.error(f'Error during storing: {e}')
            if self.mail:
                notify_error(self.plan_path,
                             'storing',
                             e,
                             run=run_number,
                             loader_config=dl_conf)


def load_plan(file_path):
    """
    Loads the plan module from the given python file.

    It loads the module and checks whether it is a valid plan module.

    Args:
        file_path: The path to the plan module.
    """
    spec = importlib.util.spec_from_file_location('plan', file_path)
    pipeline_plan = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(pipeline_plan)

    missing_members = []
    wrong_members = []
    if 'dataloader' not in dir(pipeline_plan):
        missing_members.append('dataloader')
    elif not isinstance(pipeline_plan.dataloader, Loader):
        wrong_members.append('dataloader')

    if 'pipeline' not in dir(pipeline_plan):
        missing_members.append('pipeline')

    if 'evaluator' not in dir(pipeline_plan):
        missing_members.append('evaluator')
    elif not isinstance(pipeline_plan.evaluator, Evaluator):
        wrong_members.append('evaluator')

    if 'storage_handlers' in dir(pipeline_plan):
        if isinstance(pipeline_plan.storage_handlers, Iterable):
            for handler in pipeline_plan.storage_handlers:
                if not isinstance(handler, StorageHandler):
                    wrong_members.append(f'handler: {handler}')
        elif isinstance(pipeline_plan.storage_handlers, StorageHandler):
            pipeline_plan.storage_handlers = [pipeline_plan.storage_handlers]
        else:
            wrong_members.append('storage_handler')
    else:
        pipeline_plan.storage_handlers = []

    if len(missing_members) > 0:
        raise ValueError(
            f'Module {file_path} is not a valid plan module.'
            f"The following members are missing: {','.join(missing_members)}.")

    if len(wrong_members) > 0:
        raise ValueError(f'Module {file_path} includes invalid members:'
                         f"{','.join(wrong_members)}.")

    return pipeline_plan



import unittest


class DataloaderUnitTest(unittest.TestCase):

    def _check_pair(self, x, y):
        #  data and labels must have equal lenth
        self.assertEqual(len(x), len(y))

        #  data and labels must have at least one sample
        self.assertTrue(len(x) > 0)

    def _check_train_test(self, xtrain, ytrain, xtest, ytest):
        #  at least two samples for training, at least one sample for testing
        self.assertTrue(len(xtrain) >= 2)
        self.assertTrue(len(xtest) >= 1)

        #  as many labels as training samples
        self.assertEqual(len(xtrain), len(ytrain))
        self.assertEqual(len(xtest), len(ytest))

        #  at least two distinctive classes
        self.assertTrue(len(set(ytrain)) > 1)

        #  every test class must be present for training
        self.assertTrue(set(ytrain) >= set(ytest))

    def _loader_sanity_check(
            self,
            xtrain, ytrain,
            xval=None, yval=None,
            xtest=None, ytest=None):
        self._check_pair(xtrain, ytrain)

        if xtest is not None:
            self.assertTrue(ytest is not None)
            self._check_pair(xtest, ytest)
            self._check_train_test(xtrain, ytrain, xtest, ytest)

        if xval is not None:
            self.assertTrue(yval is not None)
            self._check_pair(xval, yval)
            self._check_train_test(xtrain, ytrain, xval, yval)



"""Tool to manage data."""
import os

from logzero import logger
import yaml

from dbispipeline import utils

LINK_CONFIG_FILE = 'data/links.yaml'


def link(dataset_dir=None):
    """Links the needed datasets to the dataset folder.

    The configuration is loaded from `data/links.yaml`.

    Args:
        dataset_dir: the prefix used for the dataset paths. Overrides
            default from dbispipeline. To use the pipeline config pass None.
    """
    if dataset_dir is None:
        procject_config = utils.load_project_config()[utils.SECTION_PROJECT]
        dataset_dir = procject_config[utils.OPTION_DATASET_DIR]

        if dataset_dir == '':
            logger.error('No dataset dir is defined. Look at the README.')
            exit(4)
    logger.info('Linking datasets from: %s' % dataset_dir)

    try:
        with open(LINK_CONFIG_FILE, 'r') as yml:
            try:
                config = yaml.safe_load(yml)
            except yaml.YAMLError as ex:
                logger.exception(ex)
    except FileNotFoundError:
        logger.error('No such file or directory: %s' % LINK_CONFIG_FILE)
        exit(1)

    try:
        for dataset in config['datasets']:
            _link_dataset(dataset_dir, dataset)
    except KeyError:
        logger.error('Config does not contain datasets.')
        exit(3)


def _link_dataset(dataset_dir, dataset):
    logger.info('\t- %s' % dataset)

    src = os.path.join(dataset_dir, dataset)
    if not os.path.isdir(src):
        logger.warn('Source is not directory: %s' % src)
        return

    dst = 'data'
    if not os.path.isdir(dst):
        logger.error('No destination data directory: %s' % dst)
        exit(2)

    dataset = os.path.join(*dataset.split('/')[1:])
    dst = os.path.join(dst, dataset)

    try:
        os.symlink(src, dst, target_is_directory=True)
    except FileExistsError:
        if not os.readlink(dst) == src:
            logger.warn('Wrong source: \'%s\' instead of \'%s\'' %
                        (os.readlink(dst), src))



"""This module is called when executing from a venv mapped cli call."""
import logging
from subprocess import call

import click
import git

import dbispipeline.cli.dataset as dataset
from dbispipeline.core import Core
from dbispipeline.storage_handlers import PostGresOutputHandler
from dbispipeline.storage_handlers import PrintHandler
from dbispipeline.utils import LOGGER
from dbispipeline.utils import prepare_slurm_job
from dbispipeline.utils import restore_backup


@click.command(help='Uses the plan file specified in PLAN to run the dbis '
               'pipeline.')
@click.option('--dryrun', is_flag=True, help='Don\'t store results into DB')
@click.option('--force', is_flag=True, help='Run even if git is dirty')
@click.option('-v', '--verbose', is_flag=True, help='increase logging')
@click.option('--slurm', is_flag=True, help='create slurm file and submit')
@click.option('--restore', type=str, help='result file to be restored')
@click.option(
    '--mail',
    type=click.Choice(['none', 'run', 'total']),
    default='none',
    help='Mail notification level. Choose one of [None, \'run\', \'total\''
    ']. If set no None, no mails will be sent. if set to \'run\', one info'
    ' mail will be sent for each run. If set to \'total\', one mail will '
    'be sent after the entire pipeline is complete.')
@click.argument('plan', type=click.Path(exists=True))
def _main(dryrun, force, verbose, slurm, restore, mail, plan):
    main(dryrun, force, verbose, slurm, restore, mail, plan)


# this method is split to allow the legacy method of invoking via the modules'
# __main__.py file to re-use this code.
def main(dryrun, force, verbose, slurm, restore, mail, plan):
    """Entry point that executes the pipeline given a configuration."""
    if verbose:
        LOGGER.setLevel(logging.DEBUG)
        LOGGER.debug('setting logging level to DEBUG')

    if restore:
        restore_backup(restore, [PrintHandler(), PostGresOutputHandler()])
        exit(0)

    if not force:
        try:
            repo = git.Repo(search_parent_directories=True)
            if repo.is_dirty():
                LOGGER.error(
                    'Please commit your changes before you run the pileline.')
                exit(1)
        except git.GitError:
            pass

    if slurm:
        jobfile = prepare_slurm_job(dryrun, force, verbose, restore, mail,
                                    plan)
        call(['sbatch', jobfile])
    else:
        Core(plan, dryrun=dryrun, mail=mail).run()


@click.command(help='Links the specified dataset to the data folder.')
@click.option(
    '--dataset-dir',
    '-p',
    default=None,
    type=str,
    help='The path prefix used to find the datasets.',
)
def link(dataset_dir):
    """Links the dataset as specified in data/links.yaml."""
    dataset.link(dataset_dir=dataset_dir)



"""Module containing tools to run result analytics."""
import numpy as np
import pandas as pd

from .db import DB
from .db import DbModel
from .utils import load_project_config


def _extract_metric_results(outcome, requested_metric):
    return outcome.apply(
        lambda row: row.apply(lambda value: value[requested_metric]))


def get_results_as_dataframe(project_name,
                             table_name='results',
                             filter_git_dirty=True):
    """Returns the results stored in the database as a pandas dataframe.

    Args:
        project_name: the project name to fetch results.
        table_name: the name of the reults table.
        filter_git_dirty: defines if dirty commits are filterd.
    """
    results = pd.read_sql_table(table_name=table_name, con=DB.engine)

    if filter_git_dirty:
        results = results[results['git_is_dirty'] == False]  # noqa: E712

    return results[results['project_name'] == project_name]


def fetch_by_git_commit_id(git_commit_id):
    """Returns a query that is filtered by git commit id.

    Args:
        git_commit_id: the commit id of the returned results.
    """
    session = DB.session()
    return session.query(DbModel).filter(
        DbModel.git_commit_id == git_commit_id)


def fetch_by_row_ids(from_id, to_id=None):
    """Returns a query in the given range of ids.

    Args:
        from_id: the smalles database id included in the results of the query.
        to_id: if specified, this is the biggest database id included in the
            results of the query.
    """
    session = DB.session()
    query = session.query(DbModel).filter(DbModel.id >= from_id)
    if to_id is not None:
        return query.filter(DbModel.id <= to_id)
    return query


def fetch_by_project_name(project=None):
    """Returns a query filtered by project.

    Args:
        project: the name of the project to extract results. If None, the
            project in the dbispipeline.ini is used.
    """
    if project is None:
        project = load_project_config()['project']['name']

    session = DB.session()
    return session.query(DbModel).filter(DbModel.project_name == project)


def get_cv_epoch_evaluator_results(requested_metric=None,
                                   query_function=fetch_by_project_name):
    """Extracts CvEpochEvaluator results from the database.

    Args:
        requested_metric: allows to restrict the results to a single metric.
        query_function: a function returing a SQLAlchemy query when called.

    Returns: A tuple containing the prepared results as first element and the
        whole db entry as the second entry. The prepared results are eighter a
        pandas dataframe if a metric is requested or a dict containing a pandas
        dataframe per metric.
    """
    for run in query_function():
        if run.evaluator['class'] == 'CvEpochEvaluator':
            outcome = pd.DataFrame(run.outcome)

            if requested_metric is None:
                results = {}
                for available_metric in run.evaluator['scoring']:
                    results[available_metric] = _extract_metric_results(
                        outcome,
                        available_metric,
                    )
                yield results, run
            elif requested_metric in run.evaluator['scoring']:
                yield _extract_metric_results(outcome, requested_metric), run


def rows_to_dataframe(rows,
                      allow_git_different_rows=False,
                      allow_git_dirty_rows=False):
    """
    Converts database rows to a pandas DataFrame.

    args:
        rows: some object that iterates over rows. May be a query result or an
            actual list of rows. If this field is None or empty, an empty
            DataFrame will be returned.
        allow_git_dirty_rows: if set to true, allows that rows might have
            different git commit ids. Otherwise, an exception is thrown.
        allow_git_dirty_rows: if set to true, allows that rows have a dirty git
            status. Otherwise, an exception is thrown.

    returns: a pandas DataFrame with all columns of the database as columns.
    """
    # rows might be a query, which is not yet fetched from the database.
    if type(rows) not in [list, np.array]:
        rows = list(rows)

    if rows is None or len(rows) == 0:
        return pd.DataFrame()

    git_ids = set([row.git_commit_id for row in rows])
    if not allow_git_different_rows and len(git_ids) > 1:
        raise ValueError(f'your result contains multiple git ids: {git_ids}')

    # the DbModel objects will have additional columns that are not interesting
    # for the underlying application. In the __table__.columns field, the
    # actual list of "payload"-columns is stored.
    columns = [column.name for column in rows[0].__table__.columns]
    df_rows = []
    for row in rows:

        if not allow_git_dirty_rows and row.git_is_dirty:
            raise ValueError('your result contains dirty git rows')

        row_result = {}
        for column in columns:
            row_result[column] = getattr(row, column)
        df_rows.append(row_result)

    return pd.DataFrame(df_rows)


def _read_parameters(dictionary, prefix='', use_prefix=True):
    """Recursive helper method for extracting GridSearch param information."""
    if len(dictionary) == 0:
        return dictionary
    result = {}
    for k, v in dictionary.items():
        # parameter contains a PipelineHelper, we need recursion
        if '__selected_model' in k:
            # the [:-16] cuts off the string '__selected_model'
            key_name = prefix + k[:-16] if use_prefix else k[:-16]
            result[key_name] = str(v[0])
            result.update(_read_parameters(v[1], f'{v[0]}__'))
        else:
            key_name = prefix + k if use_prefix else k
            result[key_name] = f'{v}'
    return result


def extract_gridsearch_parameters(
    df,
    score_name,
    drop_outcome=True,
    prefix_parameter_names=True,
):
    """
    Extracts parameters from a grid search result.

    This method creates one DataFrame row for each parameter combination in the
    "outcome -> cv_results" field, and one column of each distinct parameter.
    For example, if your grid search contains a parameter `svm__C: [1, 10]`,
    then this method will add a column `C` to your DataFrame, and replace this
    row with two rows for the values 1 and 10.

    This method will recursively resolve parameters used in PipelineHelpers.

    Depending on your configurations, the output of this method may make the
    row ids no longer unique.

    before:
    row0 = {
        dataloader: XY,
        outcome: {
            cv_results: {
                params: {
                    1: { SVM__C:  1},
                    2: { SVM__C: 10},
                }
            }
        }
    }

    after:
    row0 = {
        dataloader: XY,
        SVM__C: 1,
    }
    row1 = {
        dataloader: XY,
        SVM__C: 10,
    }

    args:
        df: a pandas DataFrame object that has one column 'outcome', which
            contains dictionaries which have a field 'cv_results'. Notably,
            this is the case for the result of the rows_to_dataframe method.
        score_name: name of the field that the score should be extracted from.
            This can be a single string (e.g., 'mean_test_score') or a list of
            strings.
        drop_outcome: if true, the resulting dataframe will no longer have the
            original 'outcome' column.
        prefix_parameter_names: This parameter only affects models which have
            a PipelineHelper. If true, the resulting parameter names are
            returned by their full name. If set to false, only the part within
            the PipelineHelper is returned.
            Note that omitting this prefix may result in multiple parameters
            with the same name, possibly leading to grouping unrelated fields.

    returns: a pandas DataFrame with all possible parameters as columns, and
        all distinct parameter combinations as rows.
    """
    result_rows = []
    for _, row in df.iterrows():
        if 'outcome' not in row or 'cv_results' not in row['outcome']:
            raise ValueError(
                'this result set does not seem to contain grid '
                "search results, missing field: row['outcome']['cv_results']")
        cv = row['outcome']['cv_results']
        for combination_id, combination in cv['params'].items():
            # prevent unnecessary copy of the outcome field, which may be big
            result_row = {k: v for k, v in row.items() if k != 'outcome'}
            if not drop_outcome:
                result_row['outcome'] = row['outcome']

            if type(score_name) == str:
                score_name = [score_name]

            for score in score_name:
                result_row[score] = cv[score][combination_id]
            result_row.update(
                _read_parameters(combination,
                                 use_prefix=prefix_parameter_names))
            result_rows.append(result_row)
    return pd.DataFrame(result_rows)



"""Module containing the evaluator base class and some default evaluators."""
import datetime
import json
import os
import pickle
import warnings

import numpy as np
import pandas as pd
from sklearn import metrics
from sklearn.base import clone
from sklearn.model_selection import BaseCrossValidator
from sklearn.model_selection import GridSearchCV
from sklearn.model_selection import PredefinedSplit
from sklearn.utils.validation import indexable

from .base import Evaluator
from .utils import LOGGER


def _check_fixed_split_data_shape(data):
    assert (len(data) == 2) or (len(data) == 3)
    assert len(data[0]) == 2

    if len(data) == 2:
        assert len(data[1]) == 2
    else:
        assert len(data[2]) == 2


def _extract_train_test_data(data):
    _check_fixed_split_data_shape(data)

    if len(data) == 2:
        return data
    else:
        train, _, test = data
        return train, test


class ModelCallbackWrapper(Evaluator):
    """Allows a callback to do something with the fitted model.

    E.g. predict or store in a pickle).
    """

    def __init__(self, wrapped_evaluator, callback, *callback_args,
                 **callback_kwargs):
        """Initializes the model callback wrapper object.

        Returns the fitted model to the callback method after evaluating
        using the wrapped evluator.
        """
        if not hasattr(wrapped_evaluator, 'fitted_model'):
            raise ValueError(
                f'the wrapped evaluator {wrapped_evaluator.__class__.__name__}'
                'does not have a fitted_model attribute.')
        self.wrapped_evaluator = wrapped_evaluator
        self.callback = callback
        self.callback_args = callback_args
        self.callback_kwargs = callback_kwargs

    def evaluate(self, model, data):
        """
        Evaluates the pipeline based on the given dataset.

        Args:
            model: the model given in the pipeline.
            data: the data needed for this run, passed to grid as:
                grid.fit(*data) therefore, the data can also contain groups.

        Returns: the result of the wrapped evaluator.
        """
        result = self.wrapped_evaluator.evaluate(model, data)
        fitted_model = self.wrapped_evaluator.fitted_model
        self.callback_result_ = self.callback(fitted_model,
                                              *self.callback_args,
                                              **self.callback_kwargs)
        return result

    @property
    def configuration(self):
        """A json-like representation of the configuration."""
        return {
            'model_callback': self.callback.__name__,
            'wrapped_evaluator': self.wrapped_evaluator.__class__.__name__,
            'wrapped_configuration': self.wrapped_evaluator.configuration,
        }


class GridEvaluator(Evaluator):
    """Evalutaor that performs a grid search."""

    def __init__(self, parameters, grid_parameters):
        """
        Creates a new instance.

        Args:
            parameter: passed to the GridSearchCV as configuration.
            grid_parameter: passed to the GridSearchCV to configure the
                gridsearch.
        """
        self.parameters = parameters
        self.grid_parameters = grid_parameters
        self.model_ = None

    def evaluate(self, model, data):
        """
        Evaluates the pipline based on the given dataset.

        Args:
            model: the model given in the pipeline.
            data: the data needed for this run, passed to grid as:
                grid.fit(*data) therefore, the data can also contain groups.

        Returns: A dict containting the results of the grid search.
        """
        grid = GridSearchCV(model, self.parameters, **self.grid_parameters)
        grid.fit(*data)
        # FIXME is it intended that the default is refitting?
        if self.grid_parameters.get('refit', True):
            self.model_ = grid.best_estimator_

        try:
            best_score = grid.best_score_
            best_params = grid.best_params_
        except AttributeError:
            best_score = None
            best_params = None

        return {
            'cv_results': pd.DataFrame(grid.cv_results_).to_dict(),
            'best_score': best_score,
            'best_params': best_params,
        }

    @property
    def fitted_model(self):
        """Contains the filtered model."""
        if self.model_ is None and \
                'refit' in self.grid_parameters and \
                self.grid_parameters['refit'] is False:
            warnings.warn('you have set refit=False, but are trying to access'
                          'the fitted model. It will always be None with this'
                          'configuration.')
        return self.model_

    @property
    def configuration(self):
        """
        Returns a dict-like representation of this evaluator.

        This is for storing its state in the database.
        """
        grid_parameters = {}
        for name, param in self.grid_parameters.items():
            try:
                # this checks if the parameter can be stored in the database,
                # and falls back if it is not hashable.
                json.dumps(param)
                grid_parameters[name] = param
            except TypeError:
                grid_parameters[name] = str(param)

        return {
            'pipeline_parameters': self.parameters,
            'grid_parameters': grid_parameters,
        }


class CustomCvGridEvaluator(GridEvaluator):
    """Use this evaluator if your dataloader returns a custom cv object."""

    def __init__(self, params, grid_params):
        """Initializes the grid evaluator object.

        Args:
            params: are passed to the sklearn gridsearch.
            grid_parameters: are passed to the sklearn gridsearch.
        """
        for param in ['cv']:
            if param in grid_params:
                warnings.warn(
                    f'{param}={grid_params[param]} was set in '
                    'grid_params, but this will be overwritten by the '
                    'evaluation method. Remove this parameter to silence '
                    'this warning.', UserWarning)
        super(CustomCvGridEvaluator, self).__init__(params, grid_params)

    def evaluate(self, model, data):
        """
        Evaluates the pipline based on the given dataset.

        Args:
            model: the model given in the pipeline.
            data: the data used to load explicit train and test data.

        Returns: A dict-like object containing the results.
        """
        assert len(data) == 3
        x, y, cv = data

        if not hasattr(cv, 'split'):
            raise Exception(f'the submitted cv {cv} does not support split()')

        self.grid_parameters['cv'] = cv
        return super().evaluate(model, (x, y))


class FixedSplitGridEvaluator(GridEvaluator):
    """Runs eval on a fixed split.

    Use this evaluator if your dataloader returns explicit train / test
    samples. This evaluator will then use a PredefinedSplit to create one
    single run for each grid configuration.
    """

    def __init__(self, params, grid_params):
        """Initializes the grid evaluator object.

        Args:
            params: are passed to the sklearn gridsearch.
            grid_parameters: are passed to the sklearn gridsearch.
        """
        self.grid = None
        for param in ['cv']:
            if param in grid_params:
                warnings.warn(
                    f'{param}={grid_params[param]} was set in '
                    'grid_params, but this will be overwritten by the '
                    'evaluation method. Remove this parameter to silence '
                    'this warning.', UserWarning)
        super(FixedSplitGridEvaluator, self).__init__(params, grid_params)

    def evaluate(self, model, data):
        """
        Evaluates the pipline based on the given dataset.

        Args:
            model: the model given in the pipeline.
            data: the data used to load explicit train and test data.

        Returns: A dict containting the results of the grid search.
        """
        (xtrain, ytrain), (xtest, ytest) = _extract_train_test_data(data)
        # index=-1 means that these samples will never be used for testing
        target_indices = [-1] * len(xtrain) + [0] * len(xtest)
        self.grid_parameters['cv'] = PredefinedSplit(target_indices)

        # make sure that we have numpy-compatible data
        if type(xtrain) == list:
            xtrain = np.array(xtrain)
        if type(ytrain) == list:
            ytrain = np.array(ytrain)
        if type(xtest) == list:
            xtest = np.array(xtest)
        if type(ytest) == list:
            ytest = np.array(ytest)

        if type(xtrain) == np.ndarray:
            x = np.concatenate((xtrain, xtest))
            y = np.concatenate((ytrain, ytest))
        elif type(xtrain) == pd.DataFrame:
            x = pd.concat((xtrain, xtest))
            y = pd.concat((ytrain, ytest))
        else:
            raise ValueError(
                f'cant concatenate train and test of type {type(xtrain)}')

        return super().evaluate(model, (x, y))


class FixedSplitEvaluator(Evaluator):
    """Evaluator that runs basic classification metrics on the pipline."""

    def __init__(self, scoring):
        """
        Creates a new instance.

        Args:
            scoring: if None, then f1, accuracy, recall, precision and the
                confusion matrix are computed. Otherwise, scoring has to be a
                dict where key is the name of the score and the value contains
                a scorer.
        """
        self.scoring = scoring
        self.model_ = None

    def evaluate(self, model, data):
        """
        Evaluates the pipline based on the given dataset.

        Args:
            model: the model given in the pipeline.
            data: a tuple of (xtrain, ytrain, xtest, ytest)

        Returns: A dict containting f1, accuracy, recall, precision and the
        confusion matrix.
        """
        (xtrain, ytrain), (xtest, ytest) = _extract_train_test_data(data)
        model.fit(xtrain, ytrain)

        self.model_ = model

        result = {}
        for name, scorer in self.scoring.items():
            score = scorer(model, xtest, ytest)
            if isinstance(score, np.ndarray):
                result[name] = score.tolist()
            else:
                result[name] = score
        return result

    @property
    def fitted_model(self):
        """Contains the fitted model after training."""
        return self.model_

    @property
    def configuration(self):
        """
        Returns a dict-like representation of this evaluator.

        This is for storing its state in the database.
        """
        return {
            'scoring': list(self.scoring.keys()),
        }


class ClassificationEvaluator(FixedSplitEvaluator):
    """Evaluator that runns basic classification metrics on the pipline."""

    def __init__(self, average='macro'):
        """
        Creates a new instance.

        Args:
            average: see the sklearn doc for e.g. f1_score .
        """
        super().__init__(
            scoring={
                'f1':
                    metrics.make_scorer(metrics.f1_score, average=average),
                'accuracy':
                    metrics.get_scorer('accuracy'),
                'recall':
                    metrics.make_scorer(metrics.recall_score, average=average),
                'precision':
                    metrics.make_scorer(metrics.precision_score,
                                        average=average),
                'confusion_matrix':
                    metrics.make_scorer(metrics.confusion_matrix),
            })
        self.average = average

    @property
    def configuration(self):
        """
        Returns a dict-like representation of this evaluator.

        This is for storing its state in the database.
        """
        return {
            'average': self.average,
            **super().configuration,
        }


class EpochEvaluator(FixedSplitEvaluator):
    """Evaluator that calculates multiple scorings based on training epochs."""

    def __init__(self, scoring, scoring_step_size=10):
        """Initializes the new instance.

        Args:
            scoring: has to be a dict where key is the name of the score and
                the value contains a scorer.
            scoring_step_size: is number of epochs between each scoring call.
        """
        super().__init__(scoring=scoring)
        self.scoring_step_size = scoring_step_size

    def evaluate(self, model, data):
        """
        Evaluates the pipline based on the given dataset.

        Args:
            model: the model given in the pipeline.
            data: a tuple of (xtrain, ytrain, xtest, ytest)

        Returns: A dict-like object containing the results.
        """
        (xtrain, ytrain), (xtest, ytest) = _extract_train_test_data(data)

        try:
            model_epochs = model.epochs
        except AttributeError:
            # assume that it is a sklearn pipeline object and only the last
            # estimator supports epochs.
            model_epochs = model[-1].epochs

        assert self.scoring_step_size <= model_epochs

        trained_epochs = 0
        result = {}
        while trained_epochs < model_epochs:
            if (model_epochs - trained_epochs) < self.scoring_step_size:
                epochs = model_epochs - trained_epochs
            else:
                epochs = self.scoring_step_size
            trained_epochs += epochs
            try:
                model.fit(xtrain, ytrain, epochs=epochs)
            except ValueError:
                # assume that it is a sklearn pipeline object and only the last
                # estimator supports epochs.
                name, _ = model.steps[-1]
                fit_params = dict([(name + '__epochs', epochs)])
                model.fit(xtrain, ytrain, **fit_params)

            result[trained_epochs] = self._calc_scoring(model, xtest, ytest)

        self.model_ = model

        return result

    def _calc_scoring(self, model, xtest, ytest):
        result = {}
        for name, scorer in self.scoring.items():
            score = scorer(model, xtest, ytest)
            if isinstance(score, np.ndarray):
                result[name] = score.tolist()
            else:
                result[name] = score
        return result


class CvEpochEvaluator(Evaluator):
    """CV evaluator calculating multiple scorings based on epochs.

    For each train-test split produced by the cross-validator this evaluator
    evaluates the model after a given number of epochs using all scoring
    metrics passed as done by EpochEvaluator.
    """

    def __init__(self, cv: BaseCrossValidator, scoring, scoring_step_size=10):
        """Initializes the new instance.

        Args:
            cv: cross-validation generator that determines the cross-validation
                splitting strategy.
            scoring: has to be a dict where key is the name of the score and
                the value contains a scorer.
            scoring_step_size: is number of epochs between each scoring call.
        """
        self.scoring = scoring
        self.scoring_step_size = scoring_step_size
        self.cv = cv

    def evaluate(self, model, data):
        """
        Evaluates the pipline based on the given dataset.

        Args:
            model: the model given in the pipeline.
            data: a tuple of (data, target)

        Returns: A dict-like object containing the results.
        """
        x, y = indexable(*data)
        base_model = clone(model)

        results = {}
        for split_id, split_indices in enumerate(self.cv.split(x, y)):
            evaluator = EpochEvaluator(
                scoring=self.scoring,
                scoring_step_size=self.scoring_step_size,
            )
            train_indices, test_indices = split_indices
            split_data = (
                (x[train_indices], y[train_indices]),
                (x[test_indices], y[test_indices]),
            )
            results[split_id] = evaluator.evaluate(
                clone(base_model),
                split_data,
            )

        for epoch, scoring in _aggregate_score_dicts(results):
            try:
                results['mean'][epoch] = {}
            except KeyError:
                results['mean'] = {}
                results['mean'][epoch] = {}

            try:
                results['stdev'][epoch] = {}
            except KeyError:
                results['stdev'] = {}
                results['stdev'][epoch] = {}

            for metric, values in scoring.items():
                values = np.array(values, dtype=np.float64)
                results['mean'][epoch][metric] = np.average(
                    values,
                    axis=0,
                ).tolist()
                results['stdev'][epoch][metric] = np.std(
                    values,
                    axis=0,
                ).tolist()

        return results

    @property
    def configuration(self):
        """
        Returns a dict-like representation of this evaluator.

        This is for storing its state in the database.
        """
        return {
            'scoring': list(self.scoring.keys()),
            'scoring_step_size': self.scoring_step_size,
            'cross_validator': self.cv.__class__.__name__,
            'number_of_splits': self.cv.get_n_splits(),
        }


def _aggregate_score_dicts(scores):
    scores = list(scores.values())
    for epoch, result in scores[0].items():
        yield epoch, {
            key: np.asarray([score[epoch][key] for score in scores
                            ]) for key in result.keys()  # noqa E124
        }


def split_df_input_and_target(df, columns):
    """
    Splits a DataFrame by removing the specified columns.

    Args:
        df: The input DataFrame to split
        columns: Either a list of strings or a string denoting the columns that
            are returned as target.

    Returns:
        a tuple (x, y), where x is the DataFrame without the columns specified
        in columns, and y is either a pandas series if columns is a string,
        or y is a pandas DataFrame if columns is a list of strings.
    """
    if type(columns) == str:
        return df.drop(columns=[columns]), df[columns]
    return df.drop(columns=columns), df[columns]


class DataframeEvaluatorWrapper(Evaluator):
    """A wrapper for Evaluators for DataFrames as input."""

    def __init__(self, target_columns, wrapped_class, *args, **kwargs):
        """
        Initializes the Wrapper.

        Args:
            target_columns: a string or list of strings to use as target
                columns.
            wrapped_class: the evaluator to be wrapped.
            *args: arguments passed to the __init__ method of the wrapped class
            **kwargs: keyword args passed to the __init__ of the wrapped class
        """
        self.target_columns = target_columns
        self.wrapped_class = wrapped_class
        self.evaluator = wrapped_class(*args, **kwargs)

    def evaluate(self, model, df_data):
        """
        Evaluates the model based on a DataFrame containing all data.

        Args:
            model: the pipeline object from your plan file
            df_data: either a tuple of (train, [validate,] test) DataFrames, or
                a single DataFrame containing the data. All DataFrames must
                include all target columns with the name specified in the
                __init__ method of this evaluator.

        Returns:
            A dictionary of grid search results.
        """
        if type(df_data) == pd.DataFrame:
            result = split_df_input_and_target(df_data, self.target_columns)
        elif type(df_data) == tuple:
            result = tuple([
                split_df_input_and_target(x, self.target_columns)
                for x in df_data
            ])
        else:
            raise ValueError('unable to interpret DataFrames from data')
        return self.evaluator.evaluate(model, result)

    @property
    def configuration(self):
        """Returns the configuration of this evaluator."""
        return {
            'target_columns': self.target_columns,
            'wrapped_evaluator': self.wrapped_class.__name__,
            'wrapped_configuration': self.evaluator.configuration,
        }


def _ensure_unique_file(name, append_timestamp=True):
    if append_timestamp:
        parts = os.path.splitext(name)
        timestamp = datetime.datetime.now().strftime('%Y-%m-%d__%H-%M-%S')
        name = f'{parts[0]}_{timestamp}{parts[1]}'
    i = 0
    while os.path.isfile(name):
        parts = os.path.splitext(name)
        name = f'{parts[0]}_{i}{parts[1]}'
        i += 1
    return name


def pickle_model_callback(fitted_model, output_path):
    """Pickles fitted models, for ModelCallbackEvaluator."""
    unique_name = _ensure_unique_file(output_path)
    with open(unique_name, 'wb') as output_f:
        LOGGER.info('writing fitted model to %s', unique_name)
        pickle.dump(fitted_model, output_f)
        return unique_name



"""Utils used in the dbispipeline implementation."""
import configparser
import datetime
from email.mime.text import MIMEText
import inspect
import json
import logging
import os
from os.path import basename
from os.path import splitext
import pickle
import platform
import smtplib
import tempfile
import traceback
import warnings

import numpy as np

from . import store

logging.basicConfig(
    format='%(asctime)s %(levelname)s:%(name)-16s %(message)s',
    level=logging.INFO,
)
LOGGER = logging.getLogger('dbispipeline')

OPTION_BACKUP_DIR = 'backup_dir'
OPTION_DATASET_DIR = 'dataset_dir'

SECTION_DATABASE = 'database'
SECTION_PROJECT = 'project'
SECTION_VARIABLES = 'variables'
SECTION_MAIL = 'mail'

POSTGRES_DEFAULT_PORT = 5432


def load_project_config():
    """Loads the project configuration."""
    config_files = ['/usr/local/etc/dbispipeline.ini']
    user_home = os.getenv('HOME')
    if user_home:
        config_files.append(user_home + '/.config/dbispipeline.ini')
    config_files.append('dbispipeline.ini')
    config = configparser.ConfigParser()

    if os.getenv('DBISPIPELINE_ENV') == 'ci':
        # Sets default values if the environment is a ci
        config.read_dict({
            SECTION_DATABASE: {
                'host': 'postgres',
                'port': POSTGRES_DEFAULT_PORT,
                'user': 'runner',
                'password': 'runner_password',
                'database': 'pipelineresults',
            },
        })
        LOGGER.debug('detected CI environment')

    parsed_files = config.read(config_files)
    LOGGER.debug(f'loaded configuration from files: {parsed_files}')

    if not config.has_section(SECTION_DATABASE):
        raise KeyError('no database section found in the configuration')

    # set some default values
    if not config.has_option(SECTION_DATABASE, 'port'):
        config.set(SECTION_DATABASE, 'port', str(POSTGRES_DEFAULT_PORT))

    if not config.has_section(SECTION_PROJECT):
        config.add_section(SECTION_PROJECT)

    if not config.has_option(SECTION_PROJECT, OPTION_DATASET_DIR):
        config.set(SECTION_PROJECT, OPTION_DATASET_DIR, '')

    if os.getenv('DBISPIPELINE_DATASET_DIR'):
        dataset_dir = os.getenv('DBISPIPELINE_DATASET_DIR')
        config.set(SECTION_PROJECT, OPTION_DATASET_DIR, dataset_dir)

    if not config.has_option(SECTION_PROJECT, OPTION_BACKUP_DIR):
        config.set(SECTION_PROJECT, OPTION_BACKUP_DIR, tempfile.gettempdir())

    if os.getenv('DBISPIPELINE_BACKUP_DIR'):
        backup_dir = os.getenv('DBISPIPELINE_BACKUP_DIR')
        config.set(SECTION_PROJECT, OPTION_BACKUP_DIR, backup_dir)

    backup_dir = config[SECTION_PROJECT][OPTION_BACKUP_DIR]
    if not os.path.isdir(backup_dir):
        raise ValueError('the %s backup dir does not exist.' % backup_dir)

    return config


def get_backup_dir():
    """Returns the backup directory."""
    config = load_project_config()
    return config[SECTION_PROJECT][OPTION_BACKUP_DIR]


def load_result_backup(path):
    """
    Loads results from a pickle backup.

    Args:
        path: to the backup file
    """
    with open(path, 'rb') as f:
        return pickle.load(f)


def restore_backup(path, storage_handlers):
    """
    Reads a backup file and handles it with the given handlers.

    Args:
        path: to the backup file.
        storage_handlers: an iterable containing storage handlers
    """
    backup = load_result_backup(path)

    for handler in storage_handlers:
        handler.handle_result(backup)


def store_prediction(model, dataloader, file_name_prefix=None):
    """
    Predicts test data with fitted models.

    This method first tries to call predict() and stores the results into a
    file called <prefix>predict.npy.
    Afterwards, it tries to call predict_proba() and stores that result into
    <prefix>predict_proba.npy.


    Args:
        model: a trained model that should call predict() or predict_proba().
        dataloader: a dataloader that has the load_test() method.
        file_name_prefix: a prefix for the file names. If this prefix does not
            end with an underscore, one will be appended to it.
    """
    if not file_name_prefix:
        file_name_prefix = type(model).__name__

    if store['plan_path']:
        if file_name_prefix[-1] != '_':
            file_name_prefix += '_'
        file_name_prefix += splitext(basename(store['plan_path']))[0]

    if file_name_prefix[-1] != '_':
        file_name_prefix += '_'

    x_test, _ = dataloader.load_test()

    try:
        y_pred = model.predict(x_test)
        np.save(file_name_prefix + 'predict.npy', y_pred)
    except AttributeError:
        LOGGER.warning('Model does not support predict.')

    try:
        y_pred = model.predict_proba(x_test)
        np.save(file_name_prefix + 'predict_proba.npy', y_pred)
    except AttributeError:
        LOGGER.warning('Model does not support predict_proba.')


def notify_success(plan_path,
                   times,
                   result=None,
                   run=None,
                   subject='DBIS Pipeline: successfully finished',
                   loader_config=None):
    """
    Sends an email that new results have been successfully stored.

    Args:
        plan_path: the path to the plan file that has been used
        result: something representing a result that is tried to be included
            in the email. It should be json-dumpable.
        run: the run number. This setting is useful for multiloaders, where
            multiple runs emerge from the same plan.
        subject: email subject
        loader_config: configuration of the dataloader that was used. Useful
            for multiloaders, where multiple runs emerge from the same plan.
    """
    run_string = ''
    if run is not None:
        run_string = f' run #{run}'

    result_string = 'The results are available in the database.'
    if result and isinstance(result, dict) or isinstance(result, str):
        try:
            pretty_printed_json = json.dumps(result, indent=2, sort_keys=True)
            result_string = f'Your results are:\n\n{pretty_printed_json}'
        except Exception:
            LOGGER.warn('could not write result as pretty json string')
            pass
    computer_name = platform.node()
    duration = datetime.timedelta(seconds=int(times['eval'] - times['start']))
    message = f'''\
Hello,
your configuration file {plan_path} running on {computer_name}\
        has finished{run_string}.
The calculation took {duration}.'''

    if loader_config:
        message += f'''
The configuration of the dataloader was:
    {loader_config}'''

    message += '\n' + result_string

    _notify(message, subject)


def notify_error(plan_path,
                 error_stage,
                 error_object,
                 subject='DBIS Pipeline: error',
                 run=None,
                 loader_config=None):
    """
    Sends an email that something has gone wrong during the pipeline execution.

    Args:
        plan_path: the path to the plan file that has been used
        error_stage: string at which point the pipeline failed.
        error_object: additional information that is added in the mail body.
            should be json-dumpable.
        subject: email subject
        run: the run number. This setting is useful for multiloaders, where
            multiple runs emerge from the same plan.
        loader_config: configuration of the dataloader that was used. Useful
            for multiloaders, where multiple runs emerge from the same plan.
    """
    run_string = ''
    if run is not None:
        run_string = f' on run #{run}'
    message = f'''Hello,
unfortunately, your configuration file {plan_path} caused an error on
the pipeline running on {platform.node()} during {error_stage}{run_string}.
'''
    if loader_config:
        message += f'''
The configuration of the dataloader was:
    {loader_config}'''

    if error_object:
        message += f'''
The error message is:
    {traceback.format_tb(error_object.__traceback__)}'''

    _notify(message, subject)


def _notify(message, subject):
    cfg = load_project_config()
    if SECTION_MAIL not in cfg:
        LOGGER.debug(f'missing section [{SECTION_MAIL}]: not sending success '
                     'message: {message}')
        return

    required_fields = ['recipient', 'sender', 'smtp_server']
    for field in required_fields:
        if field not in cfg[SECTION_MAIL]:
            LOGGER.error(f'missing option in mail configuration: {field}')
            return

    # required
    recipient = cfg[SECTION_MAIL]['recipient']
    sender = cfg[SECTION_MAIL]['sender']
    host = cfg[SECTION_MAIL]['smtp_server']

    LOGGER.info(f'sending mail to {recipient}')
    msg = MIMEText(message)
    msg['Subject'] = subject
    msg['From'] = sender
    msg['To'] = recipient

    if cfg[SECTION_MAIL].getboolean('authenticate', fallback=False):

        more_required_fields = ['username', 'password']
        for field in more_required_fields:
            if field not in cfg[SECTION_MAIL]:
                LOGGER.error(f'missing field in mail configuration: {field}')
                return

        # optional
        port = cfg.getint(SECTION_MAIL, 'port', fallback=465)
        s = smtplib.SMTP_SSL(host, port)
    else:
        s = smtplib.SMTP(host)

    s.sendmail(sender, [recipient], msg.as_string())


def prepare_slurm_job(dryrun, force, verbose, restore, mail, plan):
    """
    Writes a slurm job to 'slurmjobs/<timestamp>_<jobname>.job'.

    The arguments passed to this method will be used as the arguments for the
    main method, so see those for detailed information.
    """
    config = load_project_config()
    jobname = config.get('project', 'name', fallback=None)
    if jobname is None:
        warnings.warn('no project name set, using configuration file name')
        jobname = os.path.basename(plan)
    log_directory = config.get('project', 'log_directory', fallback='logs')
    options = [
        '--partition=IFIgpu',
        f'--job-name={jobname}',
        '--account=dbis',
        '--nodes=1',
        '--tasks-per-node=1',
        f'-o {log_directory}/{jobname}.out',
        f'-e {log_directory}/{jobname}.err',
    ]

    if mail is not None:
        email = config.get('slurm', 'email', fallback=None)
        if email:
            options.append('--mail-type=ALL')
            options.append(f'--mail-user={email}')
        else:
            warnings.warn('No email address found in configuration, slurm will'
                          'not send any emails')

    content = '#!/bin/bash -l\n'
    content += '\n'.join(['#SBATCH ' + x for x in options])

    venv_manager = config.get('slurm',
                              'virtual_env_manager',
                              fallback='pipenv')
    content += f'\nsrun {venv_manager} run python -m dbispipeline {plan} '

    if dryrun:
        content += '--dryrun '
    if force:
        content += '--force '
    if restore:
        content += f'--restore {restore} '
    if verbose:
        content += '--verbose '

    if not os.path.isdir('slurmjobs'):
        os.makedirs('slurmjobs')
    timestamp = datetime.datetime.now().strftime('%Y_%m_%d__%H_%M_%S')
    plan_noext = os.path.splitext(os.path.basename(plan))[0]
    job_filename = f'{timestamp}_{jobname}_{plan_noext}.job'
    jobfile = os.path.join('slurmjobs', job_filename)
    with open(jobfile, 'w') as o_f:
        o_f.write(content)
    return jobfile


def get_job_id():
    """Returns the assigned job id.

    If run on a supported batch system (currently sge and slurm) this is the
    job id assigend by the batch system and as a fallback it is the process id.

    If DBISPIPELINE_JOB_ID is set. This overrides the other values and the
    respective values is reutrned.
    """
    # set a job id overruling all other job ids
    if os.getenv('DBISPIPELINE_JOB_ID'):
        return int(os.getenv('DBISPIPELINE_JOB_ID'))

    # slurm job id
    if os.getenv('SLURM_JOB_ID'):
        return int(os.getenv('SLURM_JOB_ID'))

    # sge job id
    if os.getenv('JOB_ID'):
        return int(os.getenv('JOB_ID'))

    # fallback job id
    return os.getpid()


def check_serializability(content):
    """
    Returns a copy of a dict object that can be json-serialized.

    For each item pair of the dictionary, if key and value are json
    serializable, the pair is kept as-is. Otherwise, the value of the pair is
    returned as string to enable storing in the database.

    Args:
        content: anything
            The object to be checked
    Returns:
        A copy of the results which can be stored in a JSON column in the db,
        or as a pickle for the pickle storage handler
    """
    try:
        _ = json.dumps(content)
        return content
    except TypeError:
        if type(content) == list or type(content) == tuple:
            cleaned = [check_serializability(x) for x in content]
        elif type(content) == dict:
            cleaned = {k: check_serializability(v) for k, v in content.items()}
        elif callable(content):
            cleaned = inspect.getsource(content).strip()
        else:
            cleaned = str(content)
        return cleaned



"""Module containing DB connectors."""
from sqlalchemy import JSON
from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import create_engine
from sqlalchemy import func
from sqlalchemy.engine.url import URL
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import sessionmaker

from .utils import LOGGER
from .utils import load_project_config

Base = declarative_base()


class ClassProperty(object):
    """Wrapper to annotate functions as classpropertyies."""

    def __init__(self, fget):
        """Initializes the wrapper."""
        self.fget = fget

    def __get__(self, cls, owner):
        """Handles the getter."""
        return self.fget.__get__(None, owner)()


class DB():
    """Database connector class."""

    _config = None
    _engine = None
    _session = None

    @ClassProperty
    @classmethod
    def engine(cls):
        """Engine property."""
        if cls._engine is None:
            cls._engine = create_engine(
                cls._db_connection_uri(),
                pool_pre_ping=True,
            )
            LOGGER.debug('connecting to database')
            Base.metadata.create_all(cls._engine)
            LOGGER.debug('connection established')
        return cls._engine

    @ClassProperty
    @classmethod
    def session(cls):
        """Session property."""
        if cls._session is None:
            cls._session = sessionmaker()
            cls._session.configure(bind=cls.engine)
        return cls._session

    @classmethod
    def _db_connection_uri(cls):
        if cls._config is None:
            cls._config = load_project_config()

        host = cls._config['database']['host']
        port = cls._config['database']['port']
        user = cls._config['database']['user']
        database = cls._config['database']['database']
        password = cls._config['database']['password']

        db_uri = URL(drivername='postgresql',
                     username=user,
                     password=password,
                     host=host,
                     port=port,
                     database=database)

        return db_uri


class DbModel(Base):
    """Model used to store the results in the db."""

    @declared_attr
    def __tablename__(cls):  # noqa N805
        try:
            config = load_project_config()
            return config['database']['result_table']
        except KeyError:
            return 'results'

    id = Column(Integer, primary_key=True)  # noqa: A003
    project_name = Column(String)
    date = Column(DateTime, default=func.now())
    git_commit_id = Column(String)
    git_remote_url = Column(String)
    git_is_dirty = Column(Boolean)
    sourcefile = Column(String)
    config = Column(String)
    dataloader = Column(JSON)
    pipeline = Column(JSON)
    evaluator = Column(JSON)
    outcome = Column(JSON)
    platform = Column(JSON)
    requirements = Column(JSON)



"""Declaring the main method of this package."""
import argparse

from dbispipeline.cli.main import main


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="executes dbis config files")
    # TODO: configurationfile is required but it is i.e. not needed for restore
    parser.add_argument('plan', help='file that holds config')
    parser.add_argument(
        '-f', '--force', action='store_true', help="Run with dirty git")
    parser.add_argument(
        '--dryrun',
        action='store_true',
        help='run without sending results to database.')
    parser.add_argument(
        '-m',
        '--mail',
        choices=[None, 'run', 'total'],
        default=None,
        help='Mail notification level. Choose one of [None, \'run\', \'total\''
        ']. If set no None, no mails will be sent. if set to \'run\', one info'
        ' mail will be sent for each run. If set to \'total\', one mail will '
        'be sent after the entire pipeline is complete.')
    parser.add_argument(
        '--restore',
        default=None,
        type=str,
        help="Restores the backup contained in the given file.")
    parser.add_argument(
        '--slurm',
        action='store_true',
        help='send this job to the slurm job queue instead of running it local')
    parser.add_argument(
        '-v', '--verbose', action='store_true', help='increase logging')
    args = parser.parse_args()
    main(
        args.dryrun,
        args.force,
        args.verbose,
        args.slurm,
        args.restore,
        args.mail,
        args.plan)



# -*- coding: utf-8 -*-

import string

from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin
"""
Provides transformers that work and modify strings
"""


class FlatteningTransformer(BaseEstimator, TransformerMixin):
    """
    Input Document Format: arbitrarily deep nested list of strings
    Output Document Format: string
    """

    def __init__(self, join_symbol=" "):
        self.join_symbol = join_symbol

    def fit(self, X, y=None):
        return self

    def _flatten(self, document):
        if type(document) != list:
            return document
        return self.join_symbol.join([self._flatten(x) for x in document])

    def transform(self, X, y=None):
        ret = []
        for document in X:
            ret.append(self._flatten(document))
        return ret


class TextCleaner(BaseEstimator, TransformerMixin):
    """
    removes all characters provided from strings.

    Input Document Format: string
    Output Document Format: string (cleaned)
    """

    def __init__(self, remove=string.punctuation):
        """
        Args:
            remove: a string containing all characters that should be removed.
                This defaults to the string.punctuation characters.
        """
        self.remove = remove

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        ret = []
        for document in X:
            ret.append(''.join([c for c in document if c not in self.remove]))
        return ret


class NgramTransformer(BaseEstimator, TransformerMixin):
    """
    input: list of strings
    output: list of ngrams found
    """

    def __init__(self, analyzer, ngram_range):
        """
        The arguments are loosely oriented to the scikit CountVectorizer
        Args:
            analyzer: 'word' or 'char'
            ngram_range: tuple
        """
        self.analyzer = analyzer
        self.ngram_range = ngram_range

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        ret = []
        for document in X:
            tmp = []
            if self.analyzer == 'word':
                words = document.split()
            elif self.analyzer == 'char':
                words = document
            for ngram_size in range(self.ngram_range[0],
                                    self.ngram_range[1] + 1):
                for i in range(len(words) - ngram_size + 1):
                    if self.analyzer == 'word':
                        tmp.append(" ".join(words[i:i + ngram_size]))
                    elif self.analyzer == 'char':
                        tmp.append("".join(words[i:i + ngram_size]))
            ret.append(tmp)
        return ret



"""Module containing functions that handle the results of an evaluator."""
import json
import os
import re

import matplotlib.pyplot as plt
from decimal import Decimal as Dec
import pandas as pd
from collections import defaultdict
import numpy as np

from .utils import LOGGER


def plot_gridsearch_parameter_gains(results,
                                    scoring='rank_test_score',
                                    invert=False):
    df = pd.DataFrame(results['cv_results']).fillna('None')
    parameter_keys = [x for x in df.columns if 'param_' in x]

    result = defaultdict(dict)
    fig, ax = plt.subplots()

    for parameter in parameter_keys:
        try:
            list(df.groupby([parameter]))
        except TypeError:
            df[parameter] = df[parameter].astype(str)
        for value, sub_df in df.groupby([parameter]):
            result[parameter][value] = sub_df[scoring].mean()

    ipos = 0
    ylabel = scoring
    for parameter, pavs in result.items():
        if len(pavs) < 2:
            LOGGER.warning(
                f'skipping less than 2 values for parameter {parameter}')
            continue
        s = sorted(pavs.items(), key=lambda x: x[1])
        values = np.array([x[1] for x in s])
        # this is required when 'rank' is selected for scoring, as a lower rank
        # means a higher performance
        if invert:
            values = 1 / values
            ylabel = f'1/{ylabel}'
        labels = []
        for label, value in s:
            ax.text(
                ipos,
                0,
                f'  {label}',
                rotation='vertical',
                ha='center',
                fontweight='bold',
                va='bottom')
            labels.append(ipos)
            ipos += 1
        ax.bar(labels, values, label=parameter)
    ax.set_ylabel(ylabel)
    ax.set_xticks([])
    ax.legend()
    plt.show()


def tex_escape(text):
    conv = {
        '&': r'\&',
        '%': r'\%',
        '$': r'\$',
        '#': r'\#',
        '_': r'\_',
        '{': r'\{',
        '}': r'\}',
        '~': r'\textasciitilde{}',
        '^': r'\^{}',
        '\\': r'\textbackslash{}',
        '<': r'\textless{}',
        '>': r'\textgreater{}',
    }
    return ''.join([conv.get(c, c) for c in str(text)])


def grid_to_latex_table(results,
                        params,
                        include_line='hline',
                        test_score_format='0.01'):
    """generates a table formatted for latex documents.

    Args:
        results: results generated by a gridsearch evaluator
        params: a list of strings containing the parameters that should be
            included (e.g., ['tfidf__ngram_range'])
        include_line: one of ['hline', 'midrule', None] to draw a line below
            the first row
    """

    if 'cv_results' not in results:
        return ''
    res = results['cv_results']
    ranks = sorted(res['rank_test_score'].items(), key=lambda x: x[1])
    columns = params + ['rank', 'test_score']
    columns = [tex_escape(c) for c in columns]
    result = '\\begin{tabular}{' + 'l' * len(columns) + '}\n'
    result += ' & '.join(columns) + '\\\\\n'
    if include_line == 'hline':
        result += '\\hline\n'
    elif include_line == 'midrule':
        result += '\\midrule\n'
    for rank in ranks:
        run_number = rank[0]
        position = rank[1]
        score = res['mean_test_score'][run_number]
        score_f = Dec(score).quantize(Dec(test_score_format))
        param_strings = []
        for p in params:
            if p not in res['params'][run_number]:
                param_strings.append('N/A')
            else:
                # these curly braces are required to prevent a square barcket
                # from being the first character in a table line, which latex
                # interprets as a parameter for a previous command
                param_strings.append(
                    '{' + tex_escape(res['params'][run_number][p]) + '}')
        params_string = ' & '.join(param_strings)
        result += f'  {params_string} & {position} & {score_f}\\\\\n'
    result += '\\end{tabular}\n'
    print()
    print(result)


def grid_to_csv_file(results,
                     filename,
                     sort_by_column='rank_test_score',
                     exclude_combined_params_field=True,
                     exclude_split_scores=False):
    """ stores grid search results into a csv file

    Args:
        results: result object generated by a gridsearch validator
        filename: where to store the resulting csv file
        sort_by_column: which column should be sorted by
        exclude_combined_params_field: in addition to each paramater passed to
            grid search, the grid search result contains a field 'params' which
            redundantly stores every parameter a second time. Set this
            parameter to False to include it in the resulting csv file.
    """

    if 'cv_results' not in results:
        LOGGER.warning(
            'grid_to_csv_file result handler could not write results, as it '
            'seemed to have received a non-gridsearch result object')
        return
    res = results['cv_results']
    df = pd.DataFrame(res).sort_values([sort_by_column])
    if exclude_combined_params_field:
        df = df.drop(['params'], axis=1)
    pat = r'split\d+_test_score'
    if exclude_split_scores:
        df = df.drop([str(x) for x in df.columns if re.match(pat, x)], axis=1)

    original_file = os.path.basename(filename)

    original_name = os.path.splitext(original_file)[0]
    original_ext = os.path.splitext(original_file)[1]
    counter = 1
    while os.path.isfile(filename):
        filename = f'{original_name}_{counter:02d}.{original_ext}'
        counter += 1

    df.to_csv(filename)
    LOGGER.info(f'wrote grid search results to {filename}.')


def print_gridsearch_results(results, include_cv_results=False):
    """Prints the results of a gridsearch.

    Args:
        results: the results passed from a gridsearch evaluator.
    """
    if include_cv_results:
        print(f'cv results:\n{json.dumps(results["cv_results"], indent=4)}')
    for key in results.keys():
        if not key == 'cv_results':
            print(f'{key}: {results[key]}')


def print_gridsearch_results_json(results):
    """Prints the results of a gridsearch.

    Args:
        results: the results passed from a gridsearch evaluator.
    """
    print(json.dumps(results['cv_results'], indent=4))


def plot_gridsearch_results(results, rank_filter='score', result_filter=None):
    """Plots the results of a gridsearch.

    Args:
        results: the results passed from a gridsearch evaluator.
        rank_filter: the measurement used to filter the results.
        result_filter: substring that must be contained in a column that the
        column is printed.
    """
    df = pd.DataFrame(results['cv_results'])

    result_cols = [name for name in df.columns if 'split' not in name]

    if result_filter:
        cols = [name for name in result_cols if result_filter in name]
    else:
        cols = result_cols
    cols.append("params")
    results = pd.DataFrame(df, columns=cols)
    results = results.rename(index=int, columns=str)
    results = results.sort_index(ascending=True)

    rank_filter = 'rank_test_' + rank_filter
    df = df[(df[rank_filter] == 1)]

    if results.shape[0] > 0:
        results.plot(kind='bar')
        plt.show()
    else:
        LOGGER.warning("No results")



"""Module containing csv dataloaders."""
from pandas import read_csv

from ..base import Loader, TrainTestLoader, TrainValidateTestLoader


class CsvLoader(Loader):
    """Loads a csv file."""

    def __init__(self, path, labels=None, **kwargs):
        """
        Creates a new instance.

        Args:
            path: the path to csv file (see pandas read_csv docs).
            labels: columns used as labels.
            kwargs: arguments that are passed to the pandas read_csv.
        """
        self.path = path
        self.labels = labels
        self.kwargs = kwargs

    def load(self):
        """
        Returns a tuple containing data and labels.

        Both are list-like if the labels argument is set. Ohterwise, the
        labels are None.
        """
        df = read_csv(filepath_or_buffer=self.path, **self.kwargs)

        if self.labels:
            return df.drop(columns=self.labels).values, df[self.labels].values

        else:
            return df.values, None

    @property
    def configuration(self):
        """Returns a dict-like representation of the configuration."""
        return {
            'path': self.path,
            'labels': self.labels,
            **self.kwargs,
        }


class CsvTrainTestLoader(TrainTestLoader):
    """Loader for train and test spilt csv."""

    def __init__(self, train_path, test_path, **kwargs):
        """
        Creates a new instance.

        Args:
            train_path: the path to train csv (see pandas read_csv docs).
            test_path: the path to test csv (see pandas read_csv docs).
            kwargs: arguments that are passed to the CsvLoader.
        """
        self.train_path = train_path
        self.test_path = test_path
        self.kwargs = kwargs

    def load_train(self):
        """Loads the data contained in the train csv file."""
        return CsvLoader(path=self.train_path, **self.kwargs).load()

    def load_test(self):
        """Loads the data contained in the test csv file."""
        return CsvLoader(path=self.test_path, **self.kwargs).load()

    @property
    def configuration(self):
        """Returns a dict-like representation of the configuration."""
        return {
            'train_path': self.train_path,
            'test_path': self.test_path,
            **self.kwargs,
        }


class CsvTrainValidateTestLoader(TrainValidateTestLoader, CsvTrainTestLoader):
    """Loader for train and test spilt csv."""

    def __init__(self, train_path, validate_path, test_path, **kwargs):
        """
        Creates a new instance.

        Args:
            train_path: the path to train csv (see pandas read_csv docs).
            validate_path: the path to validate csv (see pandas read_csv docs).
            test_path: the path to test csv (see pandas read_csv docs).
            kwargs: arguments that are passed to the CsvLoader.
        """
        self.train_path = train_path
        self.validate_path = validate_path
        self.test_path = test_path
        self.kwargs = kwargs

    def load_validate(self):
        """Loads the data contained in the validate csv file."""
        return CsvLoader(path=self.validate_path, **self.kwargs).load()

    @property
    def configuration(self):
        """Returns a dict-like representation of the configuration."""
        return {
            'train_path': self.train_path,
            'validate_path': self.validate_path,
            'test_path': self.test_path,
            **self.kwargs,
        }



# -*- coding: utf-8 -*-

from sklearn.base import BaseEstimator, TransformerMixin
"""
Contains helper classes that can be used with the scikit-learn machine
learning tools. They are primarily meant for NLP tasks, but should generalize
well.

The main contribution of these helper files are scikit pipeline elements and
the transformer picker element.
"""


class DictFieldTransformer(BaseEstimator, TransformerMixin):
    """
    Extracts a field from a dictionary object.
    """

    def __init__(self, field_name):
        """
        Args:
            field_name: the field to extract from the string
        """
        self.field_name = field_name

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        ret = []
        for document in X:
            ret.append(document[self.field_name])
        return ret


class CallbackTransformer(BaseEstimator, TransformerMixin):
    """
    Calls a predefined function on each document or on the entire data.
    """

    def __init__(self, callback, per_document=True):
        """
        Args:
            callback: the function to be called. Lambdas work well.
            per_document: if true, the function is called on each document.
                if false, the function is alled on the entirety of X in fit().
        """
        self.callback = callback
        self.per_document = per_document

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        if not self.per_document:
            return self.callback(X)
        ret = []
        for document in X:
            ret.append(self.callback(document))
        return ret



"""Module containing JSON playlist dataloader."""
import json

from ..base import Loader


class JsonPlaylistLoader(Loader):
    """Loads playlist from a json file."""

    def __init__(self, path):
        """
        Creates a new instance.

        Args:
            path: the path to the playlists dataset json file.
        """
        self.path = path

    def load(self):
        """
        Returns a tuple containing data and lables.

        Both data and labels are list-like if the labels argument is set.
        Ohterwise, the labels are None.
        """
        with open(self.path) as json_file:
            dataset = json.load(json_file)

            return dataset['data'], None

        return None, None

    def dataset_info(self):
        """Returns meta info on the dataset."""
        with open(self.path) as json_file:
            dataset = json.load(json_file)

            return dataset['info']

        return None

    @property
    def configuration(self):
        """Returns a dict-like representation of the configuration."""
        return {
            'path': self.path,
        }



"""This module provides convenience wrappers for generic loaders."""

from collections import defaultdict
import random
import warnings

import numpy as np
from sklearn.model_selection import ParameterGrid
from sklearn.preprocessing import LabelEncoder

from dbispipeline.base import Loader


class RepeatingLoader(Loader):
    """A replicator which generates repeated runs of a loader."""

    def __init__(self, loader_class, repetitions, *args, **kwargs):
        """
        A replicator which generates repeated runs of a loader.

        Args:
            loader_class: The base loader class. This class will be
                instantiated only one, and each repetition is the result of a
                call to its 'load' method.
            repetitions: How many repetitions should be produced
            args: arguments passed to the loader class' __init__
            kwargs: keyword arguments passed to the loader class' __init__
        """
        self.loader = loader_class(*args, **kwargs)
        self.repetitions = repetitions

    def load(self):
        """Returns repeated results of the loader's 'load' method."""
        for _ in range(self.repetitions):
            yield self.loader.load()

    @property
    def run_count(self):
        """Returns how many repetitions are returned by this loader."""
        return self.repetitions

    @property
    def configuration(self):
        """Returns the db-compatible configuration of this loader."""
        config = self.loader.configuration
        config.update({
            'repetitions': self.repetitions,
        })
        return config


class MultiLoaderGenerator(Loader):
    """Dynamically constructs a multiloader by combining parameters."""

    def __init__(self, loader_class, parameters):
        """
        Dynamically constructs a multiloader by combining parameters.

        Args:
            loader_class: the class of the dataloader to be instantiated. Do
                not pass an instance.
            parameters: if passed a list, this generator will return one
                dataloader instance for each entry in the list, and each entry
                in the list is passed to the constructor of the loader.
                If passed a dict, a grid of all combinations is generated and
                one loader instance is created for each combination.
        """
        self.loaders = []
        if isinstance(parameters, dict):
            for sample in ParameterGrid(parameters):
                # this produces only dicts
                self.loaders.append(loader_class(**sample))
        else:
            for sample in parameters:
                if isinstance(sample, dict):
                    self.loaders.append(loader_class(**sample))
                else:
                    self.loaders.append(loader_class(*sample))

    def load(self):
        """Loads the data of each loader constructed in the __init__ method."""
        for loader in self.loaders:
            yield loader.load()

    @property
    def configuration(self):
        """Loads the configuration of all loaders constructed in __init__."""
        for i, loader in enumerate(self.loaders):
            config = loader.configuration
            config['run_number'] = i
            config['class'] = loader.__class__.__name__
            yield config

    @property
    def run_count(self):
        """Returns how many loaders were constructed in the __init__ method."""
        return len(self.loaders)


def label_encode(cls):
    """Decorator for loaders which encodes categorical labels to integers."""
    old_load = cls.load
    old_init = cls.__init__

    def new_init(self, *args, **kwargs):
        old_init(self, *args, **kwargs)
        self.label_encoder = LabelEncoder()

    def new_load(self):
        data, targets = old_load(self)
        return data, self.label_encoder.fit_transform(targets)

    cls.__init__ = new_init
    cls.load = new_load
    return cls


def limiting(cls):
    """
    Decorator for loaders which constraints document and target numbers.

    This decorator can be used on a loader class to limit the amount of target
    classes as well as the amount of documents used for each of those classes.
    It adds the following arguments to the Loaders' constructor: max_targets,
    max_documents_per_target, strategy, random_seed.
    """
    old_load = cls.load
    old_init = cls.__init__

    def new_init(
            self,
            *args,
            max_targets=None,
            max_documents_per_target=None,
            strategy='random',
            random_seed=None,
            **kwargs):
        """
        Limiting loader which constraints document and target numbers.

        Args:
            max_targets: how many target classes should remain in the result.
                If set to None, all targets remain in the result.
            max_documents_per_target: how many documents should be left for
                each target class.
            strategy: how should the items that remain in the result be
                selected. Valid options are 'first' or 'random' (default).
            random_seed: Seed to provide to random.seed(). Ignored if strategy
                is set to 'first'.
        """
        old_init(self, *args, **kwargs)
        valid_strategies = ['first', 'random']
        if strategy not in valid_strategies:
            raise ValueError(f'the strategy {strategy} is not valid. Please '
                             f'choose from {valid_strategies}')
        if max_targets is not None and max_targets < 2:
            raise ValueError('max_targets must be 2 or greater')
        if max_documents_per_target is not None and \
                max_documents_per_target < 1:
            raise ValueError('max_document_per_target must be 1 or greater')

        self.max_targets = max_targets
        self.max_documents_per_target = max_documents_per_target
        self.strategy = strategy
        self.random_seed = random_seed

    def new_load(self):
        pairs = []
        if hasattr(cls, 'load_validate'):
            pairs.append(cls.load_validate(self))
        if hasattr(cls, 'load_test'):
            pairs.append(cls.load_test(self))
            pairs.append(cls.load_train(self))
        elif hasattr(cls, 'load'):
            pairs.append(old_load(self))
        else:
            raise ValueError('No loading methods found for this loader!')

        # first, restructure (x, y) pairs to {y: [x0, x1, ...]} dict
        # for each of the train, test, validate pairs
        dicts = []
        for pair in pairs:
            entry = defaultdict(list)
            for data, label in zip(pair[0], pair[1]):
                if isinstance(label, np.ndarray) or isinstance(label, list):
                    label = str(label)
                entry[label].append(data)
            dicts.append(entry)

        training_targets = set(dicts[-1].keys())
        selected_targets = _sample(
            values=training_targets,
            strategy=self.strategy,
            sample_limit=self.max_targets)

        result = []
        for bunch in dicts:
            bunch_result = [[], []]
            for key in selected_targets:
                values = _sample(
                    values=bunch[key],
                    strategy=self.strategy,
                    sample_limit=self.max_documents_per_target)
                for value in values:
                    bunch_result[0].append(value)
                    bunch_result[1].append(key)
            result.append(bunch_result)

        if len(result) == 1:
            return result[0]
        return result

    cls.__init__ = new_init
    cls.load = new_load
    return cls


def _sample(values, strategy, sample_limit):
    if sample_limit is None:
        return list(values)

    if sample_limit >= len(values):
        warnings.warn(
            f'sample_limit ({sample_limit}) >= targets '
            f'({len(values)})')
        return list(values)

    if strategy == 'first':
        return list(values)[:sample_limit]

    if strategy == 'random':
        return random.sample(values, sample_limit)



"""dbispipeline python packages."""
from setuptools import find_packages
from setuptools import setup

# we should not use the requirements.txt at this point after all.
# https://packaging.python.org/discussions/install-requires-vs-requirements/#requirements-files

with open('README.md') as fh:
    long_description = fh.read()

setup(
    name='dbispipeline',
    version='0.8.0',
    author='Benjamin Murauer, Michael Vötter',
    author_email='b.murauer@posteo.de',
    description='should make things more reproducible',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://git.uibk.ac.at/dbis/software/dbispipeline',
    install_requires=[
        'gitpython>2',
        'matplotlib',
        'pandas',
        'psycopg2-binary',
        'scikit-learn',
        'sqlalchemy>1.3',
        'click',
        'logzero',
        'pyyaml',
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Topic :: Scientific/Engineering',
    ],
    entry_points={
        'console_scripts': [
            'dp=dbispipeline.cli.main:_main',
            'dbispipeline=dbispipeline.cli.main:_main',
            'dbispipeline-link=dbispipeline.cli.main:link',
        ],
    },
    python_requires='>=3.6',
    include_package_data=True,
    setup_requires=[],
    test_suite='tests',
    tests_require=[],
    packages=find_packages('src'),
    package_dir={'': 'src'},
)



"""Module containing the storage handler base class and default handlers."""
import pickle
import tempfile

from .base import StorageHandler
from .db import DB
from .db import DbModel
from .utils import LOGGER
from .utils import check_serializability
from .utils import load_project_config

try:
    from pip._internal.operations import freeze
except ImportError:  # pip < 10.0
    from pip.operations import freeze


class PrintHandler(StorageHandler):
    """Prints the results."""

    def handle_result(self, result):
        """
        Prints the results of an evaluator.

        Args:
            result: passed from an evaluator.
        """
        print(result)


class PickleBackupHandler(StorageHandler):
    """Pickes the results."""

    def __init__(self, directory=None):
        """
        Intitializes the handler.

        Args:
            directory: the directory where the pickle file gets stored.
        """
        self.directory = directory

    def handle_result(self, result):
        """
        Pickles the results of an evaluator.

        Args:
            result: passed from an evaluator.
        """
        # TODO: think of a useful naming/storage strategy
        with tempfile.NamedTemporaryFile(
                dir=self.directory, suffix='.result',
                delete=False) as output_file:

            pickle.dump(result, output_file)
            LOGGER.debug(f'wrote pickle to {output_file.name}')


def _get_requirements():
    return list(freeze.freeze())


class PostGresOutputHandler(StorageHandler):
    """Stores the meta-data, configuration and results of a run in the DB."""

    def __init__(self, table='results'):
        """
        Creates a new instance.

        Args:
            table: the name of the database table used to store the results.
        """
        self.config = load_project_config()
        self.engine = DB.engine
        self.session = DB.session

    def handle_result(self, result):
        """
        Stores the result of an evaluator in the DB.

        Args:
            result: passed from an evaluator.
        """
        result = check_serializability(result)
        commit_id, remote_url, is_dirty = result['git_info']
        if 'project' in self.config and 'name' in self.config['project']:
            project_name = self.config['project']['name']
        else:
            project_name = result['sourcefile']
        db_entry = DbModel(
            project_name=project_name,
            git_commit_id=commit_id,
            git_remote_url=remote_url,
            git_is_dirty=is_dirty,
            sourcefile=result['sourcefile'],
            config=result['plan_content'],
            dataloader=result['dataloader'],
            pipeline=result['pipeline'],
            evaluator=result['evaluator'],
            outcome=result['outcome'],
            platform=result['platform'],
            requirements=_get_requirements())

        session = self.session()

        session.add(db_entry)
        session.commit()
